# coding=utf-8
#
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip import BaseManagement as BigipBaseManagement
from f5.iworkflow.cm import Cm
from f5.iworkflow.resource import PathElement
from f5.iworkflow.shared import Shared
from f5.iworkflow.tm import Tm
from f5.sdk_exception import F5SDKError
from icontrol.session import iControlRESTSession

import re


class BaseManagement(object):
    def __init__(self, hostname, username, password, **kwargs):
        self.args = self.parse_arguments(
            hostname, username, password, **kwargs
        )
        self.icrs = self._get_icr_session(**self.args)
        self.configure_meta_data(**self.args)
        self.set_icr_metadata(self.icrs)

    def parse_arguments(self, *args, **kwargs):
        result = dict(
            timeout=kwargs.pop('timeout', 30),
            port=kwargs.pop('port', 443),
            icontrol_version=kwargs.pop('icontrol_version', ''),
            verify=kwargs.pop('verify', False),
            token=kwargs.pop('token', False)
        )
        if kwargs:
            raise TypeError('Unexpected **kwargs: %r' % kwargs)
        result.update(dict(
            hostname=args[0],
            username=args[1],
            password=args[2]
        ))
        return result

    def _get_icr_session(self, *args, **kwargs):
        return iControlRESTSession(
            username=kwargs['username'],
            password=kwargs['password'],
            timeout=kwargs['timeout'],
            verify=kwargs['verify'],
            token=kwargs['token']
        )

    def configure_meta_data(self, *args, **kwargs):
        self._meta_data = {
            'allowed_lazy_attributes': [Tm, Cm, Shared],
            'hostname': kwargs['hostname'],
            'port': kwargs['port'],
            'device_name': None,
            'local_ip': None,
            'bigip': self,
            'icontrol_version': kwargs['icontrol_version'],
            'username': kwargs['username'],
            'password': kwargs['password'],
            'tmos_version': None,
        }

    def set_icr_metadata(self, icrs):
        self._meta_data['icr_session'] = icrs


class RegularManagementRoot(BaseManagement, PathElement):
    def __init__(self, hostname, username, password, **kwargs):
        super(RegularManagementRoot, self).__init__(
            hostname, username, password, **kwargs
        )
        self.set_metadata_uri(**self.args)
        self.post_configuration_setup()

    def set_metadata_uri(self, *args, **kwargs):
        self._meta_data['uri'] = 'https://{0}:{1}/mgmt/'.format(
            kwargs['hostname'], kwargs['port']
        )

    def post_configuration_setup(self):
        self._get_os_version()

    @property
    def hostname(self):
        return self._meta_data['hostname']

    @property
    def icontrol_version(self):
        return self._meta_data['icontrol_version']

    @property
    def tmos_version(self):
        if self._meta_data['tmos_version'] is None:
            self._get_os_version()
        return self._meta_data['tmos_version']

    def _get_os_version(self):
        connect = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['uri'] + \
            'shared/identified-devices/config/device-info'
        response = connect.get(base_uri)
        version = response.json()['version']
        self._meta_data['tmos_version'] = version


class ManagementProxy(object):
    def __new__(cls, *args, **kwargs):
        proxy_to = kwargs.pop('proxy_to', None)
        device_group = kwargs.pop('device_group', 'cm-cloud-managed-devices')

        mgmt = ManagementRoot(args[0], args[1], args[2], **kwargs)
        uuid = cls._get_identifier(mgmt, proxy_to)
        if uuid is None:
            raise F5SDKError(
                "The specified device was missing a UUID. "
                "This should not happen!"
            )
        bigip = BigipBaseManagement(
            mgmt.args['hostname'],
            mgmt.args['username'],
            mgmt.args['password'],
            port=mgmt.args['port']
        )
        bigip.icrs = mgmt.icrs
        uri = ''.join([
            'https://{0}:{1}/mgmt/shared/resolver/device-groups/',
            '{2}/devices/{3}/rest-proxy/mgmt/'
        ])
        bigip._meta_data['uri'] = uri.format(
            mgmt.args['hostname'],
            mgmt.args['port'],
            device_group,
            uuid
        )
        bigip.post_configuration_setup()
        return bigip

    @staticmethod
    def _get_identifier(mgmt, proxy_to):
        if proxy_to is None:
            raise F5SDKError(
                "An identifier to a device to proxy to must be provided."
            )

        if re.search(r'([0-9-a-z]+\-){4}[0-9-a-z]+', proxy_to, re.I):
            return proxy_to
        return ManagementProxy._get_device_uuid(mgmt, proxy_to)

    @staticmethod
    def _get_device_uuid(mgmt, proxy_to):
        dg = mgmt.shared.resolver.device_groups
        collection = dg.cm_cloud_managed_devices.devices_s.get_collection(
            requests_params=dict(
                params="$filter=hostname+eq+'{0}'&$select=uuid".format(
                    proxy_to
                )
            )
        )
        if len(collection) > 1:
            raise F5SDKError(
                "More that one managed device was found with this hostname. "
                "Proxied devices must be unique."
            )
        elif len(collection) == 0:
            raise F5SDKError(
                "No device was found with that hostname"
            )
        else:
            resource = collection.pop()
            return resource.pop('uuid', None)


class ManagementRoot(BaseManagement, PathElement):
    """An interface to a single BIG-IP"""
    def __new__(cls, *args, **kwargs):
        proxy_to = kwargs.pop('proxy_to', None)
        device_group = kwargs.pop('device_group', 'cm-cloud-managed-devices')
        if proxy_to:
            return ManagementProxy(
                proxy_to=proxy_to, device_group=device_group, *args, **kwargs
            )
        else:
            return RegularManagementRoot(*args, **kwargs)



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® cluster device submodule

REST URI
    ``http://localhost/mgmt/tm/cm/device/``

GUI Path
    ``Device Management --> Devices``

REST Kind
    ``tm:cm:device:*``
"""


from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Devices(Collection, CommandExecutionMixin):
    """BIG-IP® cluster devices collection.

    """
    def __init__(self, cm):
        super(Devices, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [Device]
        self._meta_data['attribute_registry'] =\
            {'tm:cm:device:devicestate': Device}
        self._meta_data['allowed_commands'].append('mv')


class Device(Resource):
    """BIG-IP® cluster device object.

    """
    def __init__(self, device_s):
        super(Device, self).__init__(device_s)
        self._meta_data['required_json_kind'] = 'tm:cm:device:devicestate'
        self._meta_data['required_creation_parameters'].update(('partition',))



# coding=utf-8
#
# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® iApp (application) module

REST URI
    ``http://localhost/mgmt/sys/application/``

GUI Path
    ``iApps``

REST Kind
    ``tm:sys:application:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource

from requests import HTTPError


class Application(OrganizingCollection):
    """BIG-IP® iApp collection."""
    def __init__(self, sys):
        super(Application, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Aplscripts,
            Customstats,
            Services,
            Templates
        ]


class Aplscripts(Collection):
    """BIG-IP® iApp script collection."""
    def __init__(self, application):
        super(Aplscripts, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Aplscript]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:apl-script:apl-scriptstate': Aplscript}


class Aplscript(Resource):
    """BIG-IP® iApp script resource."""
    def __init__(self, apl_script_s):
        super(Aplscript, self).__init__(apl_script_s)
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:apl-script:apl-scriptstate'


class Customstats(Collection):
    """BIG-IP® iApp custom stats sub-collection."""
    def __init__(self, application):
        super(Customstats, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Customstat]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:custom-stat:custom-statstate': Customstat}


class Customstat(Resource):
    """BIG-IP® iApp custom stats sub-collection resource."""
    def __init__(self, custom_stat_s):
        super(Customstat, self).__init__(custom_stat_s)
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:custom-stat:custom-statstate'


class Services(Collection):
    """BIG-IP® iApp service sub-collection."""
    def __init__(self, application):
        super(Services, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Service]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:service:servicestate': Service}


class Service(Resource):
    """BIG-IP® iApp service sub-collection resource"""
    def __init__(self, service_s):
        super(Service, self).__init__(service_s)
        self._meta_data['required_creation_parameters'].update(
            ('template', 'partition')
        )
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:service:servicestate'
        self._meta_data['disallowed_load_parameters'] = \
            set(['template', 'trafficGroup'])

    def _create(self, **kwargs):
        '''Create service on device and create accompanying Python object.

        :params kwargs: keyword arguments passed in from create call
        :raises: HTTPError
        :returns: Python Service object
        '''

        try:
            return super(Service, self)._create(**kwargs)
        except HTTPError as ex:
            if "The configuration was updated successfully but could not be " \
                    "retrieved" not in ex.response.text:
                raise

            # BIG-IP® will create in Common partition if none is given.
            # In order to create the uri properly in this class's load,
            # drop in Common as the partition in kwargs.
            if 'partition' not in kwargs:
                kwargs['partition'] = 'Common'
            # Pop all but the necessary load kwargs from the kwargs given to
            # create. Otherwise, load may fail.
            kwargs_copy = kwargs.copy()
            for key in kwargs_copy:
                if key not in self._meta_data['required_load_parameters']:
                    kwargs.pop(key)
            # If response was created successfully, do a local_update.
            # If not, call to overridden _load method via load
            return self.load(**kwargs)

    def update(self, **kwargs):
        '''Push local updates to the object on the device.

        :params kwargs: keyword arguments for accessing/modifying the object
        :returns: updated Python object
        '''

        inherit_device_group = self.__dict__.get('inheritedDevicegroup', False)
        if inherit_device_group == 'true':
            self.__dict__.pop('deviceGroup')
        return self._update(**kwargs)

    def _load(self, **kwargs):
        '''Load python Service object with response JSON from BIG-IP®.

        :params kwargs: keyword arguments for talking to the device
        :returns: populated Service object
        '''
        # Some kwargs should be popped before we do a load
        for key in self._meta_data['disallowed_load_parameters']:
            if key in kwargs:
                kwargs.pop(key)

        self._check_load_parameters(**kwargs)
        name = kwargs.pop('name', '')
        partition = kwargs.pop('partition', '')
        read_session = self._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']

        load_uri = self._build_service_uri(base_uri, partition, name)
        response = read_session.get(load_uri, uri_as_parts=False, **kwargs)
        return self._produce_instance(response)

    def _build_service_uri(self, base_uri, partition, name):
        '''Build the proper uri for a service resource.

        This follows the scheme:
            <base_uri>/~<partition>~<<name>.app>~<name>

        :param base_uri: str -- base uri for container
        :param partition: str -- partition for this service
        :param name: str -- name of the service
        :returns: str -- uri to access this service
        '''
        name = name.replace('/', '~')
        return '%s~%s~%s.app~%s' % (base_uri, partition, name, name)

    def exists(self, **kwargs):
        '''Check for the existence of the named object on the BIG-IP

        Override of resource.Resource exists() to build proper URI unique to
        service resources.

        Sends an HTTP GET to the URI of the named object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it returns :obj:`True`.

        For any other errors are raised as-is.

        :param kwargs: Keyword arguments required to get objects
        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: bool -- The objects exists on BIG-IP® or not.
        :raises: :exc:`requests.HTTPError`, Any HTTP error that was not status
                 code 404.
        '''

        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = False
        session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        partition = kwargs.pop('partition')
        name = kwargs.pop('name')

        exists_uri = self._build_service_uri(base_uri, partition, name)
        kwargs.update(requests_params)
        try:
            session.get(exists_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True


class Templates(Collection):
    """BIG-IP® iApp template sub-collection"""
    def __init__(self, application):
        super(Templates, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Template]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:template:templatestate': Template}


class Template(Resource):
    """BIG-IP® iApp template sub-collection resource"""
    def __init__(self, template_s):
        super(Template, self).__init__(template_s)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:template:templatestate'



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® System Software Hotfix sub-module

REST URI
    ``http://localhost/mgmt/tm/sys/software/hotfix``

GUI Path
    ``System --> Software Management --> Hotfix List``

REST Kind
    ``tm:sys:software:hotfix*``
"""

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedOperation


class Hotfix_s(Collection, CommandExecutionMixin):
    """BIG-IP® system software hotfix collection."""
    def __init__(self, software):
        super(Hotfix_s, self).__init__(software)
        self._meta_data['allowed_lazy_attributes'] = [Hotfix]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:software:hotfix:hotfixstate': Hotfix}
        self._meta_data['allowed_commands'].append('install')
        self._meta_data['required_command_parameters'].update((
            'name', 'volume'))


class Hotfix(Resource):
    """BIG-IP® system software hotfix resource."""
    def __init__(self, images):
        super(Hotfix, self).__init__(images)
        self._meta_data['required_json_kind'] = \
            'tm:sys:software:hotfix:hotfixstate'

    def create(self, **kwargs):
        """Create is not supported for hotfix resource.

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method." % self.__class__.__name__
        )

    def modify(self, **kwargs):
        """Modify is not supported for hotfix resource.

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method." % self.__class__.__name__
        )

    def update(self, **kwargs):
        """Update is not supported for hotfix resource.

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the update method." % self.__class__.__name__
        )



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IQ® Device Groups (shared) module for CM AutoDeployed Devices

REST URI
    ``http://localhost/mgmt/shared/resolver/device-groups/cm-bigip-allBigIpDevices``

REST Kind
    ``shared:resolver:device-groups:*``
"""

from f5.bigiq.resource import Collection
from f5.bigiq.resource import OrganizingCollection
from f5.bigiq.resource import Resource
from f5.sdk_exception import F5SDKError


class Cm_BigIp_AllBigIpDevices(OrganizingCollection):
    def __init__(self, device_groups):
        super(Cm_BigIp_AllBigIpDevices, self).__init__(device_groups)
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:devicegroupstate'
        self._meta_data['attribute_registry'] = {
            'shared:resolver:device-groups:devicegroupdevicecollectionstate':
                Devices_s
        }
        self._meta_data['allowed_lazy_attributes'] = [
            Devices_s
        ]

    def _format_resource_name(self):
        return 'cm-bigip-allBigIpDevices'


class Devices_s(Collection):
    def __init__(self, cm_bigip_allbigipdevices):
        super(Devices_s, self).__init__(cm_bigip_allbigipdevices)
        self._meta_data['allowed_lazy_attributes'] = [Device]
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:devicegroupdevicecollectionstate'
        self._meta_data['attribute_registry'] = {
            'shared:resolver:device-groups:restdeviceresolverdevicestate': Device  # NOQA
        }


class Device(Resource):
    def __init__(self, devices_s):
        super(Device, self).__init__(devices_s)
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:restdeviceresolverdevicestate'
        self._meta_data['required_creation_parameters'] = set(
            ('address', 'password', 'rootPassword')
        )
        self._meta_data['required_load_parameters'] = str(('uuid',))

    def update(self, **kwargs):
        raise F5SDKError(
            'Auto Deploy items can be created or deleted, not updated'
        )



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® cluster module

REST URI
    ``http://localhost/mgmt/tm/cm/``

GUI Path
    ``Device Management``

REST Kind
    ``tm:cm:*``
"""

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.cm.device import Devices
from f5.bigip.tm.cm.device_group import Device_Groups
from f5.bigip.tm.cm.failover_status import Failover_Status
from f5.bigip.tm.cm.sync_status import Sync_Status
from f5.bigip.tm.cm.traffic_group import Traffic_Groups
from f5.bigip.tm.cm.trust import Add_To_Trust
from f5.bigip.tm.cm.trust import Remove_From_Trust
from f5.bigip.tm.cm.trust_domain import Trust_Domains


class Cm(OrganizingCollection, CommandExecutionMixin):
    """BIG-IP® Cluster Organizing Collection."""
    def __init__(self, cm):
        super(Cm, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [
            Devices,
            Device_Groups,
            Traffic_Groups,
            Trust_Domains,
            Failover_Status,
            Sync_Status,
            Add_To_Trust,
            Remove_From_Trust,
        ]
        self._meta_data['allowed_commands'].append('run')



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils.version import LooseVersion
from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedOperation


class Blocking_Settings(UnnamedResource):
    """BIG-IP® ASM Blocking-Settings Unnamed Resource."""
    def __init__(self, policy):
        super(Blocking_Settings, self).__init__(policy)
        tmos_v = policy._meta_data['bigip']._meta_data['tmos_version']
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [
            Evasions_s, Http_Protocols_s, Violations_s,
            Web_Services_Securities_s
        ]

        if LooseVersion(tmos_v) < LooseVersion('12.0.0'):
            # kind has changed in v12, hence we need the below
            self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:blocking-settingcollectionstate'
        else:
            self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings'

        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:blocking-settings:evasions:evasioncollectionstate': Evasions_s,
            'tm:asm:policies:blocking-settings:violations:violationcollectionstate': Violations_s,
            'tm:asm:policies:blocking-settings:http-protocols:http-protocolcollectionstate': Http_Protocols_s,
            'tm:asm:policies:blocking-settings:web-services-securities:web-services-securitycollectionstate': Web_Services_Securities_s
        }


class Evasions_s(Collection):
    """BIG-IP® ASM Evasions sub-collection."""
    def __init__(self, blocking_settings):
        super(Evasions_s, self).__init__(blocking_settings)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Evasion]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:evasions:evasioncollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:blocking-settings:evasions:evasionstate': Evasion
        }


class Evasion(AsmResource):
    """BIG-IP® ASM Evasions resource."""
    def __init__(self, evasions_s):
        super(Evasion, self).__init__(evasions_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:evasions:evasionstate'
        self._meta_data['reduction_forcing_pairs'] = []

    def create(self, **kwargs):
        """Create is not supported for Evasion resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for Evasion resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )


class Violations_s(Collection):
    """BIG-IP® ASM Violations sub-collection"""
    def __init__(self, blocking_settings):
        super(Violations_s, self).__init__(blocking_settings)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Violation]
        self._meta_data['required_json_kind'] = \
            'tm:asm:policies:blocking-settings:violations:violationcollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:blocking-settings:violations:violationstate':
                Violation
        }


class Violation(AsmResource):
    """BIG-IP® ASM Violations resource."""
    def __init__(self, violations_s):
        super(Violation, self).__init__(violations_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:violations:violationstate'
        self._meta_data['reduction_forcing_pairs'] = []

    def create(self, **kwargs):
        """Create is not supported for Violation resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for Violation resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )


class Http_Protocols_s(Collection):
    """BIG-IP® ASM Http-Protocols sub-collection"""
    def __init__(self, blocking_settings):
        super(Http_Protocols_s, self).__init__(blocking_settings)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Http_Protocol]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:http-protocols:http-protocolcollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:blocking-settings:http-protocols:http-protocolstate': Http_Protocol
        }


class Http_Protocol(AsmResource):
    """BIG-IP® ASM Http-Protocols  resource."""
    def __init__(self, http_protocols_s):
        super(Http_Protocol, self).__init__(http_protocols_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:http-protocols:http-protocolstate'
        self._meta_data['reduction_forcing_pairs'] = []

    def create(self, **kwargs):
        """Create is not supported for Http Protocol resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for Http Protocol resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )


class Web_Services_Securities_s(Collection):
    """BIG-IP® ASM Web-Services-Securities sub-collection"""
    def __init__(self, blocking_settings):
        super(Web_Services_Securities_s, self).__init__(blocking_settings)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = \
            [Web_Services_Security]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:web-services-securities:web-services-securitycollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:blocking-settings:web-services-securities:web-services-securitystate': Web_Services_Security
        }


class Web_Services_Security(AsmResource):
    """BIG-IP® ASM Web-Services-Securities resource."""
    def __init__(self, web_serices_securities_s):
        super(Web_Services_Security, self).__init__(
            web_serices_securities_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:blocking-settings:web-services-securities:web-services-securitystate'
        self._meta_data['reduction_forcing_pairs'] = []

    def create(self, **kwargs):
        """Create is not supported for Web Services Security resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for Web Services Security resources

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )



# coding=utf-8
#
# Copyright 2018 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Directory: net module: timer-policy.

REST URI
    ``https://localhost/mgmt/tm/net/timer-policy``

GUI Path
    ``Network --> Service Policies --> Timer Policies``

REST Kind
    ``tm:net:timer-policy:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Timer_Policys(Collection):
    def __init__(self, net):
        super(Timer_Policys, self).__init__(net)
        self._meta_data['minimum_version'] = '12.0.0'
        self._meta_data['allowed_lazy_attributes'] = [Timer_Policy]
        self._meta_data['attribute_registry'] = {
            'tm:net:timer-policy:timer-policystate': Timer_Policy
        }


class Timer_Policy(Resource):
    def __init__(self, timer_policys):
        super(Timer_Policy, self).__init__(timer_policys)
        self._meta_data['minimum_version'] = '12.0.0'
        self._meta_data['required_json_kind'] = "tm:net:timer-policy:timer-policystate"



# coding=utf-8
#
# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® net module

REST URI
    ``http://localhost/mgmt/tm/net/``

GUI Path
    ``Network``

REST Kind
    ``tm:net:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.net.arp import Arps
from f5.bigip.tm.net.bwc import Bwc
from f5.bigip.tm.net.dns_resolver import Dns_Resolvers
from f5.bigip.tm.net.fdb import Fdb
from f5.bigip.tm.net.interface import Interfaces
from f5.bigip.tm.net.route import Routes
from f5.bigip.tm.net.route_domain import Route_Domains
from f5.bigip.tm.net.selfip import Selfips
from f5.bigip.tm.net.service_policy import Service_Policys
from f5.bigip.tm.net.timer_policy import Timer_Policys
from f5.bigip.tm.net.trunk import Trunks
from f5.bigip.tm.net.tunnels import TunnelS
from f5.bigip.tm.net.vlan import Vlans


class Net(OrganizingCollection):
    def __init__(self, tm):
        super(Net, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Arps,
            Bwc,
            Dns_Resolvers,
            Fdb,
            Interfaces,
            Routes,
            Route_Domains,
            Selfips,
            Service_Policys,
            Timer_Policys,
            Trunks,
            TunnelS,
            Vlans
        ]



# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

from functools import wraps
import time

from f5.utils.util_exceptions import UtilError


class MaximumAttemptsReached(UtilError):
    def __init__(self, *args, **kwargs):
        super(MaximumAttemptsReached, self).__init__(*args, **kwargs)


def poll_for_exceptionless_callable(callable, attempts, interval):
    '''Poll with a given callable for a specified number of times.

    :param callable: callable to invoke in loop -- if no exception is raised
                    the call is considered succeeded
    :param attempts: number of iterations to attempt
    :param interval: seconds to wait before next attempt
    '''

    @wraps(callable)
    def poll(*args, **kwargs):
        for attempt in range(attempts):
            try:
                return callable(*args, **kwargs)
            except Exception as ex:
                if attempt == attempts-1:
                    raise MaximumAttemptsReached(ex)
                time.sleep(interval)
                continue
    return poll
